#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <memory.h>

#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <kqml.h>

#if (! defined(ERR)) || (ERR != -1)
#	undef ERR
#	define ERR -1
#endif

#if (! defined(TRUE)) || (! TRUE)
#	undef TRUE
#	define TRUE 1
#endif

#if (! defined(FALSE)) || (FALSE)
#	undef FALSE
#	define FALSE 0
#endif

static void Usage (void);
static void ListAgents (void);

void main(int argc, char *argv[])
{
	char *buffer, *temp;
	char *progFunc;
	int i, size, len;
	kqml_message *reply_msg;

	if (! (progFunc = strrchr(argv[0], (int)'/')))
		progFunc = argv[0];
	else
		progFunc++;

	if (strcmp (progFunc, "agentadd") == 0)
	{
		if ( ((argc - 1) / 3) * 3 != (argc - 1) )
		{
			Usage ();
			exit (1);
		}

		if (kqml_initialize ("admin-agent-add", KQML_NOT_AN_ANS) <= -1)
		{
			fprintf (stderr, "kqml_initialize failed\n");
			exit (1);
		}

		size = 0;
		buffer = NULL;
		for (i = 1; i < argc; i += 3)
		{
			len = strlen(argv[i]) + strlen(argv[i + 2]) + 4;
			if (size < len)
			{
				size = len;
				if (buffer)
				{
					buffer = realloc (buffer, size);
				}
				else
				{
					buffer = malloc (size);
				}
			}
			if (buffer == NULL)
			{
				fprintf (stderr, "Memory allocation error.\n");
				exit (1);
			}

			temp = buffer;
			*temp++ = '(';
			temp = memccpy (temp, argv[i], 0, len);
			*(temp - 1) = ' ';
			temp = memccpy (temp, argv[i + 2], 0, len);
			*(temp - 1) = ')';
			*temp = '\0';

			if (kqml_send_msg (KQML_SEND_TIMEOUT, "register", buffer, 
				kqml_ans_name, "t", &reply_msg, ":ontology", argv[i + 1], 
				NULL) == 1)
			{
				printf ("Registered agent %s proto %s @ %s\n",	
					argv[i], argv[i + 1], argv[i + 2]);
			}
			else
			{
				printf ("Registration error\n");
			}
		}
		if (buffer)
		{
			free (buffer);
		}
		ListAgents ();
	}
	else if (strcmp(progFunc, "agentls") == 0)
	{
		if (kqml_initialize ("admin-agent-ls", KQML_NOT_AN_ANS) <= -1)
		{
			fprintf (stderr, "kqml_initialize failed\n");
			exit (1);
		}

		ListAgents ();
	} 
	else if (strcmp(progFunc, "agentrm") == 0)
	{
		if ( ((argc - 1) / 2) * 2 != (argc - 1) )
		{
			Usage ();
			exit (1);
		}

		if (kqml_initialize ("admin-agent-rm", KQML_NOT_AN_ANS) <= -1)
		{
			printf ("kqml_initialize failed\n");
			exit (1);
		}

		for (i = 1; i < argc; i += 2)
		{
			if (kqml_send_msg (KQML_SEND_TIMEOUT, "unregister", argv[i], 
					kqml_ans_name, NULL, NULL, ":ontology", argv[i + 1], NULL) == 1)
			{
				printf ("Removed %s\n", argv[i]);
			}
			else
			{
				printf ("Could not send remove message for %s\n", argv[i]);
			}
		}

		ListAgents ();
	}
	else if (strcmp(progFunc, "anscheck") == 0)
	{
		int sd, port;
		struct sockaddr_in name;
		unsigned long addr_type;
		char *addr, *host, *port_str;
		struct hostent *host_info;

		addr = NULL;
		if (argc != 2)
		{
			addr = getenv ("KQML_ANS");
		}
		else
		{
			addr = argv[1];
		}

		if (addr == NULL)
		{
			fprintf (stderr, "anscheck: KQML_ANS not defined!\n");
			exit (2);
		}

		host = addr;
		port_str = strchr (addr, ':');
		if (port_str == NULL)
		{
			fprintf (stderr, "anscheck: KQML_ANS format error!\n");
			exit (2);
		}
		*port_str = '\0';
		port_str++;
		port = atoi (port_str);
		if (port == 0)
		{
			exit (2);
		}

		sd = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
		if (sd <= -1)
		{
			exit (2);
		}

		if (isdigit (host[0]))
		{
			if ((addr_type = inet_addr (host)) == INADDR_BROADCAST)
			{
				exit (2);
			}
			host_info = gethostbyaddr ((char *)&addr_type, sizeof (addr_type), 
				AF_INET);
		}
		else
		{
			host_info = gethostbyname (host);
		}

		if (host_info == NULL)
		{
			exit (2);
		}
			
		memset (&name, 0, sizeof (name));
		name.sin_family = AF_INET;
		name.sin_port = htons (port);
		memcpy (&name.sin_addr, host_info->h_addr_list[0], host_info->h_length);

		if (connect (sd, &name, sizeof (name)) <= -1)
		{
			exit (1);
		}

		close (sd);
	}
	else if (strcmp(progFunc, "anskill")== 0)
	{
		if (kqml_initialize ("admin-agent-kill", KQML_NOT_AN_ANS) <= -1)
		{
			printf ("kqml_initialize failed\n");
			exit (1);
		}

		kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(shutdown)", 
				kqml_ans_name, "t", NULL, NULL);
		printf ("ANS might have been terminated\n");
	}
	else
	{
		Usage();
		exit(1);
	}
	exit (0);
}

static void Usage(void)
{
	printf("Usage: agentadd <agentName> <protocol> <addr_info> ... |\n");
	printf("       agentls                                         |\n");
	printf("       agentrm <agentName> <protocol> ...              |\n");
	printf("       anscheck [addr_info]                            |\n");
	printf("       anskill\n");
}

static void ListAgents()
{
	kqml_message *reply_msg;
	char *buffer, *temp, *name, *protocol, *addr_info;

	reply_msg = NULL;

	if (kqml_send_msg (KQML_SEND_TIMEOUT, "ask-all", "*", kqml_ans_name, 
			"t", &reply_msg, ":ontology", "*", NULL) <= -1)
	{
		fprintf (stderr, "Query Error.\n");
		return ;
	}

	if (kqml_stricmp (kqml_get_performative (reply_msg), "reply") != 0)
	{
		fprintf (stderr, "Received a [%s] from the ANS.\n",
			kqml_get_performative (reply_msg));
		return ;
	}

	buffer = kqml_get_field (":content", reply_msg, NULL, NULL);
	if (buffer == NULL)
	{
		fprintf (stderr, "Empty Message from ANS.\n");
		return ;
	}
	
	temp = strchr (buffer, '(');
	if (temp == NULL)
	{
		fprintf (stderr, "Message format error (initial) [%s].\n", buffer);
		return ;
	}
	temp++;

	while (1)
	{
		while (*temp)
		{
			if (*temp == ')' || *temp == '(')
			{
				break;
			}
			else
			{
				temp++;
			}
		}

		if (*temp == ')')
		{
			temp++;
			break;
		}
		else if (*temp != '(')
		{
			fprintf (stderr, "Message format error (name) [%s].\n", 
				buffer);
			return ;
		}
		else
		{
			temp++;
		}

		name = strtok (temp, " \t\n");
		if (name == NULL)
		{
			fprintf (stderr, "Message format error (name 2) [%s].\n", buffer);
			return ;
		}

		temp = strchr (temp, 0);
		temp++;

		printf ("%s (", name);

		while (1)
		{
			while (*temp)
			{
				if (*temp == ')' || *temp == '(')
				{
					break;
				}
				else
				{
					temp++;
				}
			}

			if (*temp == ')')
			{
				temp++;
				break;
			}
			else if (*temp != '(')
			{
				fprintf (stderr, "Message format error (protocol) [%s].\n", 
					buffer);
				return ;
			}
			else
			{
				temp++;
			}

			protocol = strtok (temp, " \t\n");
			if (protocol == NULL)
			{
				fprintf (stderr, "Message format error (protocol 2) [%s].\n", 
					buffer);
				return ;
			}

			addr_info = strtok (NULL, ")");
			if (addr_info == NULL)
			{
				fprintf (stderr, "Message format error (addr_info) [%s].\n", 
					buffer);
				return ;
			}

			printf ("%s %s) ", protocol, addr_info);

			temp = strchr (addr_info, 0);
			temp++;
		}

		printf ("\n");
	}

	kqml_free_msg (reply_msg);
}
