#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <ctype.h>

#include <kqml.h>

#include <kqml/kqmllib.h>

static int check_expression (char *string, int len);
int _kqml_check_word (char *string, int len);
static int check_string (char *string, int len);
static int check_byte (char *string, int len);
static int check_quotation (char *string, int len);
static int check_comma (char *string, int len);
static int check_nested (char *string, int len);
static int check_for_string_wrap (char *string, int len);

/* 
	ret_len is length after ignoring leading and trailing spaces
	ret_string points to the beginning of the new string

	Returns the type 
*/
kqml_expression_type
_kqml_check_type (char *string, int len, char **ret_string, int *ret_len)
{
	char *temp;
	int ret_value;

	/* Skip over the leading spaces */
	while (len > 0 && isspace(*string))
	{
		string++;
		len--;
	}

	/* New starting point */
	if (ret_string)
	{
		*ret_string = string;
	}

	/* Ignore the trailing spaces */
	temp = string + len - 1;
	while (len > 0 && isspace(*temp))
	{
		temp--;
		len--;
	}

	if (ret_len)
	{
		*ret_len = len;
	}

	/* After Ignoring leading and trailing spaces, if nothing left */
	if (len == 0)
	{
		return KQML_TYPE_STRING;
	}
	
	ret_value = check_expression (string, len);

	if (ret_value != 0)
	{
		if (check_for_string_wrap (string, len))
			return KQML_TYPE_STRING;
		else
			return KQML_TYPE_BYTE;
	}
	else
	{
		switch (*string)
		{
		case '\'':
			return KQML_TYPE_QUOTED;
		case '`':
			return KQML_TYPE_COMMA;
		case '"':
			if (len > 1024)
			{
				if (*ret_string)
				{
					*ret_string = string + 1;
				}
				if (*ret_len)
				{
					*ret_len = len - 2;
				}
				return KQML_TYPE_BYTE;
			}
			else
				return KQML_TYPE_STRING;
		case '#':
			return KQML_TYPE_BYTE;
		case '(':
			return KQML_TYPE_SEXPR;
		default:
			return KQML_TYPE_WORD;
		}

		return KQML_TYPE_UNKNOWN ; /* NOT REACHED */
	}
}

static int
check_for_string_wrap (char *string, int len)
{
	if (len > 1022)
	{
		return FALSE;
	}

	while (len > 0)
	{
		if (*string == '\\')
		{
			len--;
			string++;
		}
		else if (*string == '"')
		{
			len = -1;
		}
		else if (! isprint(*string))
		{
			len = -1;
		}
		string++;
		len--;
	}

	if (len != 0)
		return FALSE;
	else
		return TRUE;
}

static int 
check_expression (char *string, int len)
{
	switch (*string)
	{
	case '\'':
		return check_quotation (string, len);
		break;
	case '`':
		return check_comma (string, len);
		break;
	case '"':
		return check_string (string, len);
		break;
	case '#':
		return check_byte (string, len);
		break;
	case '(':
		return check_nested (string, len);
		break;
	default:
		return _kqml_check_word (string, len);
		break;
	}
	
	return -1; /* Not reached */
}

int 
_kqml_check_word (char *string, int len)
{
	static char inited = FALSE;
	static char specialTable[256];

	if (! inited)
	{
		char *str1, *str2;

		str1 = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
		str2 = "<>=+-*/&^~_@$%:.!?";
		
		memset(specialTable, 0, sizeof(specialTable));
		while (*str1)
		{
			specialTable[(int)*str1] = TRUE;
			str1++;
		}
		while (*str2)
		{
			specialTable[(int)*str2] = TRUE;
			str2++;
		}
		inited =TRUE;
	}

	while (len > 0 && specialTable[(int)*string])
	{
		string++;
		len--;
	}
	
	return len;
}

static int 
check_string (char *string, int len)
{
	string++;
	len--;

	while (len > 0)
	{
		if (*string == '\\')
		{
			len--;
			string++;
		}
		else if (*string == '"')
		{
			len--;
			break;
		}
		else if (! isprint(*string))
		{
			break;
		}
		string++;
		len--;
	}

	if (*string != '"')
		return -1;
	else
		return len;
}

static int 
check_byte (char *string, int len)
{
	int byte_len;
	char *dummy;

	byte_len = strtol (string+1, &dummy, 10);
	if (*dummy != '"' || byte_len < 0)
	{
		return -1;
	}

	return (len - byte_len - (dummy - string) - 1);
}

static int 
check_quotation (char *string, int len)
{
	return check_expression (string + 1, len - 1);
}

static int 
check_comma (char *string, int len)
{
	return check_expression (string + 1, len - 1);
}

static int 
check_nested (char *string, int len)
{
	int ret_len;

	string++;
	len--;

	/* Skip white space after ( and before the first expression */
	while (len > 0 && isspace(*string))
	{
		string++;
		len --;
	}

	if (len == 0 || *string == ')')
		return -1;

	do
	{
		ret_len = check_expression (string, len);

		if (ret_len != -1)
		{
			string = string + len - ret_len;
			
			/* The next character should be a ')' or a space */
			if (*string != ')' && !isspace(*string))
				ret_len = -1;
		}

		len = ret_len;
			
		/* Skip white space between expressions */
		while (len > 0 && isspace (*string))
		{
			string++;
			len--;
		}
	}
	while (len > 0 && *string != ')');

	if (len <= 0)
	{
		return -1;
	}

	/* Definitely the current char is ')' */
	/* The trailing spaces skipped by check_type or check_nested*/
	return len-1;
}

#ifdef TESTING

main ()
{
	char string[256];
	char *new_string;
	int i, new_len;

	while (fgets(string, sizeof(string), stdin))
	{
		i = check_type (string, strlen(string), &new_string, &new_len);
		printf ("[%.*s] Len=%d\n Type=%s", new_len, new_string, new_len,
			(i == KQML_TYPE_WORD ? "kqml_word" :
				(i == KQML_TYPE_STRING ? "kqml_string" :
				(i == KQML_TYPE_BYTE ? "kqml_byte" :
				(i == KQML_TYPE_QUOTED  ? "kqml_quoted" :
				(i == KQML_TYPE_COMMA ? "kqml_comma" :
				(i == KQML_TYPE_SEXPR ? "kqml_nested" : "kqml_unknown")))))));
	}
}

#endif
