:-module(pkril, [kqmlInitialize/1, kqmlInitialize/2, kqmlRegister/3, 
	kqmlInfo/4, kqmlAgentName/1, kqmlANSName/1, kqmlAddress/2,
	kqmlSendMsg/3, kqmlSendMsg/4, kqmlDeliverMsg/2, 
	kqmlPrintMsg/1, kqmlPrintMsg/2, 
	kqmlGetPerformative/2, kqmlGetField/5, kqmlGetIthField/6, 
	kqmlGetFieldCount/2,
 	kqmlBuildMsg/2, kqmlBuildMsg/4, 
	kqmlPutField/4, kqmlPutFieldList/4, kqmlPutEmbeddedMsg/4,
 	kqmlParse/2, kqmlParse/3, kqmlMsgToStr/2, kqmlStrToTerm/2, kqmlTermToStr/2,
	kqmlStrICmp/2, kqmltolowerAtom/2, kqmltolowerStr/2, kqmltolowerChar/2,
 	kqmlInterpreter/0]).

:-dynamic (kqmlInfo/4).			%% kqmlInfo(AgentName, ANSName, Proto, Addr)
:-dynamic (handlerTable/3).	%% handlerTable(Performative, Predicate, Module)
:-dynamic (streamInfo/1).		%% streamInfo(ConversionStream)
:-dynamic (ansHandlerTable/2).%% ansHandlerTable(Module, Predicate)

dbglevelDB('KQML_DBG_ALWAYS', 0).
dbglevelDB('KQML_DBG_WARNING', 3).
dbglevelDB('KQML_DBG_INFORM', 5).
dbglevelDB('KQML_DBG_CHATTER', 7).
dbglevelDB('KQML_DBG_VERBOSE', 9).
dbglevelDB(_, 1). 

foreign_file('$KQML_HOME/lib/libspwrap.a', ['C_kqmlInit', 'C_kqmlRegister', 
	'C_kqmlSend', 'C_kqmlDeliver', 'C_kqmlDebug', 
	'C_kqmlParse', 'C_kqmlSprintMsg']).

foreign('C_kqmlInit', c, c_kqmlInit(+integer, +string, -term, -string,
	-string, -string, -string, [-integer])).
foreign('C_kqmlRegister', c_kqmlRegister(+string, +integer, [-integer])).
foreign('C_kqmlSend', c_kqmlSend(+integer, +integer, +term, -term, [-integer])).
foreign('C_kqmlDeliver', c_kqmlDeliver(+term, [-integer])).
foreign('C_kqmlParse', c_kqmlParse(+integer, +term, -term, -term, [-integer])).
foreign('C_kqmlSprintMsg', c_kqmlSprintMsg(+term, -term, [-integer])).
foreign('C_kqmlDebug', c_kqmlDebug(+integer, +string)).

%% _____________________________________________________________________
%% load the foreign files 
%% _____________________________________________________________________
:- load_foreign_files('$KQML_HOME/lib/libspwrap.a', 
		['$KQML_HOME/lib/libkqml.a', '-ll']).

kqmlInitialize(Name):-
	kqmlInitialize(Name, nil).

kqmlInitialize(Name, AnsFunc):-
	%% _____________________________________________________________________
	%% make sure that agent name is an atom
	%% _____________________________________________________________________
	!, 
	abolish(foreign_file/2),
	abolish(foreign/2),
	(atom(Name) ->
		true;
		(format(user_error, "Agent Name should be an atom~n", []),
		fail)),

	(AnsFunc == nil ->
		Flag is 0;
		Flag is 1),

	c_kqmlInit(Flag, Name, StreamPtr, AgentName, ANSName, Protocol, AddrInfo, 1),
	stream_code(StreamHandle, StreamPtr),

	(Module:Pred = AnsFunc; (Module = user, Pred = AnsFunc)),
	retractall(ansHandlerTable(_, _)),
	asserta(ansHandlerTable(Module, Pred)),

	retractall(kqmlInfo(_, _, _, _)),
	asserta(kqmlInfo(AgentName, ANSName, Protocol, AddrInfo)),
	retractall(streamInfo(_)),
	asserta(streamInfo(StreamHandle)).

kqmlAgentName(X):-
	kqmlInfo(X, _, _, _).
kqmlANSName(X):-
	kqmlInfo(_, X, _, _).
kqmlAddress(Protocol, AddrInfo):-
	kqmlInfo(_, _, Protocol, AddrInfo).

kqmlRegister(Predicate, Performative, Blocking):-
	!, retractall(handlerTable(Performative, _, _)),
	(Module:Pred = Predicate; (Module = user, Pred = Predicate)),
	asserta(handlerTable(Performative, Pred, Module)),
	c_kqmlRegister(Performative, Blocking, 1).

kqmlPrintMsg(Msg):-
	kqmlPrintMsg(user_output, Msg).

kqmlPrintMsg(Stream, Msg):-
	ground(Msg), Msg \== nil,
	kqmlGetPerformative(Msg, Perf),
	format(Stream, "Performative: ~a~n", [Perf]),
	print_keyword_value(Stream, Msg, 1).

kqmlGetPerformative(Msg, Perf):-
	ground(Msg), Msg \== nil,
	[Perf | _] = Msg;
	Perf = nil.

kqmlGetField(Field, Msg, Value, ValueLen, ValueType):-
	ground(Msg), Msg \== nil,
	(member((Field, Value, ValueLen, ValueType), Msg)->
		true;
		(Value = nil,
		ValueLen = nil,
		ValueType = nil)).

kqmlGetFieldCount(Msg, FieldCount):-
	ground(Msg), Msg \== nil,
	length(Msg, MsgLen),
	Count is MsgLen - 1,
	FieldCount = Count.

kqmlGetIthField(Index, Msg, Field, Value, ValueLen, ValueType):-
	kqmlGetFieldCount(Msg, MaxField),
	Index > 0, Index =< MaxField,
	nthElement(Index, Msg, (Field, Value, ValueLen, ValueType)).

kqmlBuildMsg(Perf, Msg):-
	kqmlBuildMsg(Perf, Msg, [], []).

kqmlBuildMsg(Perf, [Perf | Msg], FieldList, ValueList):-
	buildMsg(Msg, FieldList, ValueList).

buildMsg([(Field, Value, ValueLen, 0) | NewMsg], [Field | FieldList],
	[IValue | ValueList]):-
	atom_to_list(IValue, Value),
	length(Value, ValueLen),
	buildMsg(NewMsg, FieldList, ValueList).
buildMsg([], [], []).

kqmlPutField([Field | FieldList], Msg, [Value | ValueList], NewMsg):-
	kqmlPutField(Field, Msg, Value, TempMsg),
	kqmlPutField(FieldList, TempMsg, ValueList, NewMsg).
kqmlPutField([], Msg, [], Msg).

kqmlPutField(Field, Msg, IValue, NewMsg):-
	ground(Msg), Msg \== nil,
	atom_to_list(IValue, Value),
	length(Value, ValueLen),
	(memberchk((Field, OldValue, OldValueLen, OldValueType), Msg) 	->
		(ValueLen \== 0 ->
			substitute((Field, OldValue, OldValueLen, OldValueType), Msg, 
				(Field, Value, ValueLen, 0), NewMsg);
			delete(Msg, (Field, OldValue, OldValueLen, OldValueType), NewMsg));
		(ValueLen \== 0,
		append(Msg, [(Field, Value, ValueLen, 0)], NewMsg))).

kqmlPutEmbeddedMsg(Field, Msg, EmbedMsg, NewMsg):-
	kqmlMsgToStr(EmbedMsg, Str),
	kqmlPutField(Field, Msg, Str, NewMsg).

kqmlMsgToStr(Msg, String):-
	c_kqmlSprintMsg(Msg, String, 1).

kqmlSendMsg(Msg, ReplyMsg, RetValue):-
	kqmlSendMsg(-1, Msg, ReplyMsg, RetValue).

kqmlSendMsg(Timeout, Msg, ReplyMsg, RetValue):-
	ReplyMsg \== nil, 
	c_kqmlSend(1, Timeout, Msg, ReplyMsg, RetValue),
	(RetValue == 0; RetValue == 1).

kqmlSendMsg(Timeout, Msg, nil, RetValue):-
	!, 
	c_kqmlSend(0, Timeout, Msg, _, RetValue),
	(RetValue == 0; RetValue == 1).

kqmlDeliverMsg(Msg, RetValue):-
	!, 
	c_kqmlDeliver(Msg, RetValue),
	(RetValue == 0; RetValue == 1).

kqmlParse(String, Msg):-
	kqmlParse(String, _, Msg).

kqmlParse(String, LeftOver, Msg):-
	!, 
	length(String, StringLen),
	c_kqmlParse(StringLen, String, LeftOver, Msg, 1),
	kqmlPrintMsg(user_error, Msg).

kqmlStrToTerm(String, Term):-
	streamInfo(StreamHandle),
	flush_output(StreamHandle),
	%% Writing whitespace, . ,whitespace to avoid read from hanging
%        format("~n BAD STREAM !!!~n",[]),
	format(StreamHandle, "~s . ", [String]),
%        format("~n BAD STREAM !!! -after ~n",[]),
	read(StreamHandle, Term).

kqmlTermToStr(Term, String):-
	streamInfo(StreamHandle),
	flush_output(StreamHandle),
	format(StreamHandle, "~q", [Term]),
	readstr(StreamHandle, String).


kqmlStrICmp([], []).

kqmlStrICmp([Head1 | Tail1], [Head2 | Tail2]):-
	tolowerChar(Head1, Head),
	tolowerChar(Head2, Head),
	kqmlStrICmp(Tail1, Tail2).


print_keyword_value(Stream, Msg, Index):-
	kqmlGetIthField(Index, Msg, Field, Value, ValueLen, ValueType),
	format(Stream, "   (~a, ~s, ~D ~d)~n", 
		[Field, Value, ValueLen, ValueType]),
	NextIndex is Index + 1,
	print_keyword_value(Stream, Msg, NextIndex).

print_keyword_value(Stream, _, _):-
	format(Stream, "~n", []),
	flush_output(Stream).

readstr(StreamHandle, String):-
	readstr_(StreamHandle, String, []).
readstr_(StreamHandle, String, Acc):-
	getchar(StreamHandle, X),
	append(Acc, [X], Acc2),
	readstr_(StreamHandle, String, Acc2).
readstr_(_, String, String):- !.

getchar(StreamHandle, X):-
	get0(StreamHandle, X),
	X \== -1,
	X \== 0.

kqmlInterpreter:-
	ttyflush,
	on_exception(Except, read(X), pkril:exception(Except)),
	((X \== end_of_file, X \== halt) ->
		(on_exception(Except, call(user:X), pkril:exception(Except)),
		kqmlInterpreter);
		halt).

kqmlInterpreter:-
	kqmlInterpreter.

ansHandler(Name, LookupAgent, Proto, AddrInfo):-
	ansHandlerTable(Module, Predicate),
	Function =.. [Predicate, Name, LookupAgent, Proto, AddrInfo],
	on_exception(Except, call(Module:Function),
		pkril:ansException(Except, LookupAgent, Proto, AddrInfo)).

handler(Content, Msg, ReplyMsg):-
	kqmlGetPerformative(Msg, Performative),
	(handlerTable(Performative, Predicate, Module)->
		true;
		(handlerTable([], Predicate, Module)->
		true;
		(Module = pkril, Predicate = errorHandler))),
	Function =.. [Predicate, Content, Msg, ReplyMsg],
	on_exception(Except, call(Module:Function), 
		pkril:handlerException(Except, ReplyMsg)),
	(ground(ReplyMsg) ->
		true;
		ReplyMsg == nil).

errorHandler(_, _, ReplyMsg):-
	kqmlBuildMsg('error', Temp1),
	kqmlPutField(':content', Temp1, "Not ready for performative yet", Temp2),
	ReplyMsg = [Temp2].

ansException(Except, _, _, _):-
	exception(Except).
ansException(_, Name, Proto, AddrInfo):-
	Name = nil,
	Proto = nil,
	AddrInfo = nil.

handlerException(Except, _):-
	exception(Except).
handlerException(_, ReplyMsg):-
	kqmlBuildMsg('error', Temp1),
	kqmlPutField(':content', Temp1, "SPKRIL: Internal S/W Error", Temp2),
	ReplyMsg = [Temp2].

exception(Except):-
	print_message(error, Except),
	!,
	fail.

dbgmsg(DebugLevelAtom, Message):-
	dbglevelDB(DebugLevelAtom, DebugLevel),
	c_kqmlDebug(DebugLevel, Message).

%% utility functions. 
%% imported the list functions from the library to avoid dependancy on
%% the lists library
kqmltolowerAtom(Atom1, Atom2):-
	name(Atom1, String1),
	kqmltolowerStr(String1, String2),
	name(Atom2, String2).


kqmltolowerStr([], []).

kqmltolowerStr([Head1 | Tail1], [Head2 | Tail2]):-
	kqmltolowerChar(Head1, Head2),
	kqmltolowerStr(Tail1, Tail2).


kqmltolowerChar(Char1, Char2):-
	((Char1 >= "A", Char1 =< "Z") ->
		Char2 is Char1 - "A" + "a";
		Char2 is Char1).

atom_to_list(Atom, List):-
	(atom(Atom) ->
		name(Atom, List);
		List = Atom).

append([], L1, L1).
append([Head | L1], L2, [Head | L3]):-
	append(L1, L2, L3).

delete([], _, []).
delete([Head|Tail], Element, Rest) :-
   Head==Element, !,
   delete(Tail, Element, Rest).
delete([Head|Tail], Element, [Head|Rest]) :-
   delete(Tail, Element, Rest).

member(Element, [Head|Tail]):-
	member_(Tail, Head, Element).

member_(_, Element, Element).
member_([Head|Tail], _, Element):-
	member_(Tail, Head, Element).

memberchk(Element, [Element|_]):- !.
memberchk(Element, [_|Rest]) :-
	memberchk(Element, Rest).

substitute(_, [], _, []):- !.
substitute(OldElem, [OldHead|OldRest], NewElem, [NewElem|NewRest]) :-
	OldElem==OldHead, !,
	substitute(OldElem, OldRest, NewElem, NewRest).
substitute(OldElem, [NotElem|OldRest], NewElem, [NotElem|NewRest]):-
	substitute(OldElem, OldRest, NewElem, NewRest).

nthElement(0, [Head | _], Head):- !.
nthElement(Index, [_ | Tail], Element):-
	NewIndex is Index - 1,
	nthElement(NewIndex, Tail, Element).
